package com.googlecode.alfresco.repository.query;

import java.util.Date;

import org.alfresco.service.cmr.repository.NodeRef;
import org.alfresco.service.namespace.QName;

import com.googlecode.alfresco.repository.query.impl.Field;

/**
 * Defines operations for building {@link Query} instances. Use a {@link QueryBuilderFactory} to create instances.
 *
 * @author Laurens Fridael
 *
 */
public interface QueryBuilder {

	// Left-hand side values.

	public QueryBuilder property(QName property);

	public QueryBuilder field(Field field);

	public QueryBuilder field(String field);

	// Conditions and right-hand side values.

	public QueryBuilder matches(String value);

	public QueryBuilder matches(Date value);

	public QueryBuilder matches(QName value);

	public QueryBuilder matches(Variable variable);

	public QueryBuilder startsWith(String value);

	public QueryBuilder startsWith(Variable variable);

	public QueryBuilder contains(String value);

	public QueryBuilder contains(Variable variable);

	public QueryBuilder isChildOf(NodeRef nodeRef);

	public QueryBuilder isChildOf(Variable variable);

	/**
	 * Matches against a given node type and any subtypes.
	 *
	 * @throws IllegalArgumentException
	 *             If type is null.
	 */
	public QueryBuilder isOfType(QName type);

	public QueryBuilder isOfType(Variable variable);

	/**
	 * Matches node aspects against any element of a given set of aspects.
	 *
	 * @params aspects
	 * @throws IllegalArgumentException
	 *             If aspects or any of its elements is null or empty.
	 */
	public QueryBuilder isOfAnyTypeOf(QName... types);

	/**
	 * Matches against a given aspect.
	 *
	 * @param aspect
	 * @throws IllegalArgumentException
	 *             If aspect is null.
	 */
	public QueryBuilder hasAspect(QName aspect);

	public QueryBuilder hasAspect(Variable variable);

	/**
	 * Matches node aspects against any element of a given set of aspects.
	 *
	 * @param aspects
	 * @throws IllegalArgumentException
	 *             If aspects or any of its elements is null or empty.
	 */
	public QueryBuilder hasAnyAspectOf(QName... aspects);

	/**
	 * Matches node aspects against all given aspects.
	 *
	 * @throws IllegalArgumentException
	 *             If aspects or any of its elements is null or empty.
	 */
	public QueryBuilder hasAllAspectsOf(QName... aspects);

	/**
	 * Precedes a new condition that uses a logical AND operation.
	 *
	 * @return
	 */
	public QueryBuilder and();

	/**
	 * Precedes a new condition that uses a logical OR operation.
	 *
	 * @return
	 */
	public QueryBuilder or();

	// Query creation.

	/**
	 * Creates the {@link Query} represented by this {@link QueryBuilder}.
	 *
	 * @return
	 */
	public Query createQuery();

	/**
	 * Starts a subquery.
	 *
	 * @return The QueryBuilder representing the subquery.
	 */
	public QueryBuilder startSubquery();

	/**
	 * Ends a subquery. Checks for matching calls to {@link #startSubquery()} and throws an
	 * {@link IllegalArgumentException} If there is mismatch between number of calls.
	 *
	 * @return The {@link QueryBuilder} for the containing (parent) query.
	 * @throws IllegalStateException
	 *             If there is no parent query.
	 */
	public QueryBuilder endSubquery();

	public QueryBuilder subquery(Query subquery);

	public QueryBuilder subquery(QueryBuilder subquery);
}