package com.googlecode.alfresco.repository.query.impl;

import java.util.Date;

import org.alfresco.service.cmr.repository.NodeRef;
import org.alfresco.service.namespace.QName;
import org.springframework.util.Assert;

import com.googlecode.alfresco.repository.query.Query;
import com.googlecode.alfresco.repository.query.QueryBuilder;
import com.googlecode.alfresco.repository.query.Variable;

/**
 * Default {@link QueryBuilder} implementation.
 *
 * @author Laurens Fridael
 *
 */
public class QueryBuilderImpl implements QueryBuilder {

	private final QueryBuilder parent;

	/**
	 * The query being populated.
	 */
	private final Query query = new Query();

	/**
	 * The current logical operation.
	 */
	private LogicalOperation logicalOperation = null;

	/**
	 * The left-hand side of the current condition.
	 */
	private Object leftHandSide = null;

	protected QueryBuilderImpl() {
		this(null);
	}

	protected QueryBuilderImpl(final QueryBuilder parent) {
		this.parent = parent;
	}

	protected QueryBuilder getParent() {
		return parent;
	}

	protected Query getQuery() {
		return query;
	}

	protected void startCondition(final Object leftHandSide) {
		if (getQuery().isEmpty() == false) {
			Assert.state(logicalOperation != null,
					"No next logical operation available. Make sure you put logical operations between conditions.");
			getQuery().addTerm(logicalOperation);
		}
		logicalOperation = null;
		this.leftHandSide = leftHandSide;
	}

	protected void endCondition(final Operator operator, final Object rightHandSide) {
		Assert.state(leftHandSide != null, "No left-hand side available");
		getQuery().addTerm(new Condition<Object, Object>(leftHandSide, operator, rightHandSide));
	}

	protected void addCondition(final Object condition) {
		if (getQuery().isEmpty() == false) {
			Assert.state(logicalOperation != null,
					"No next logical operation available. Make sure you put logical operations between conditions.");
			getQuery().addTerm(logicalOperation);
		}
		getQuery().addTerm(condition);
		logicalOperation = null;
	}

	@Override
	public Query createQuery() {
		return new Query(getQuery());
	}

	@Override
	public QueryBuilder property(final QName property) {
		Assert.notNull(property, "Property cannot be null.");
		startCondition(property);
		return this;
	}

	@Override
	public QueryBuilder field(final Field field) {
		Assert.notNull(field);
		startCondition(field);
		return this;
	}

	@Override
	public QueryBuilder field(final String field) {
		Assert.notNull(field);
		startCondition(field);
		return this;
	}

	@Override
	public QueryBuilder matches(final String value) {
		Assert.hasText(value, "Value cannot be empty.");
		endCondition(Operator.MATCHES, value);
		return this;
	}

	@Override
	public QueryBuilder matches(final QName value) {
		Assert.notNull(value, "Value cannot be null.");
		endCondition(Operator.MATCHES, value);
		return this;
	}

	@Override
	public QueryBuilder matches(final Date value) {
		Assert.notNull(value, "Value cannot be null.");
		endCondition(Operator.MATCHES, value);
		return this;
	}

	@Override
	public QueryBuilder matches(final Variable variable) {
		Assert.notNull(variable, "Variable cannot be null.");
		endCondition(Operator.MATCHES, variable);
		return this;
	}

	@Override
	public QueryBuilder startsWith(final String value) {
		Assert.hasText(value, "Value cannot be empty.");
		endCondition(Operator.STARTS_WITH, value);
		return this;
	}

	@Override
	public QueryBuilder startsWith(final Variable variable) {
		Assert.notNull(variable, "Variable cannot be null.");
		endCondition(Operator.STARTS_WITH, variable);
		return this;
	}

	@Override
	public QueryBuilder contains(final String value) {
		Assert.hasText(value, "Value cannot be empty.");
		endCondition(Operator.CONTAINS, value);
		return this;
	}

	@Override
	public QueryBuilder contains(final Variable variable) {
		Assert.notNull(variable, "Variable cannot be null.");
		endCondition(Operator.CONTAINS, variable);
		return this;
	}

	@Override
	public QueryBuilder isChildOf(final NodeRef nodeRef) {
		Assert.notNull(nodeRef, "NodeRef cannot be null.");
		addCondition(new Condition<Field, NodePath>(Field.PATH, Operator.STARTS_WITH, new NodePath(nodeRef)));
		return this;
	}

	@Override
	public QueryBuilder isChildOf(final Variable variable) {
		Assert.notNull(variable, "Variable cannot be null.");
		addCondition(new Condition<Field, Variable>(Field.PATH, Operator.STARTS_WITH, variable));
		return null;
	}

	@Override
	public QueryBuilder isOfType(final QName type) {
		Assert.notNull(type, "Type cannot be null.");
		addCondition(new Condition<Field, QName>(Field.TYPE, Operator.MATCHES, type));
		return this;
	}

	@Override
	public QueryBuilder isOfType(final Variable variable) {
		Assert.notNull(variable, "Variable cannot be null.");
		addCondition(new Condition<Field, Variable>(Field.TYPE, Operator.MATCHES, variable));
		return this;
	}

	@Override
	public QueryBuilder isOfAnyTypeOf(final QName... types) {
		Assert.notEmpty(types, "Types cannot be empty.");
		for (int i = 0, n = types.length; i < n; i++) {
			final QName type = types[i];
			Assert.notNull(type, "Type cannot be null.");
			addCondition(new Condition<Field, QName>(Field.TYPE, Operator.MATCHES, type));
			if (i < n - 1) {
				or();
			}
		}
		return this;
	}

	@Override
	public QueryBuilder hasAspect(final QName aspect) {
		Assert.notNull(aspect, "Aspect cannot be null.");
		addCondition(new Condition<Field, QName>(Field.ASPECT, Operator.MATCHES, aspect));
		return this;
	}

	@Override
	public QueryBuilder hasAspect(final Variable variable) {
		Assert.notNull(variable, "Variable cannot be null.");
		addCondition(new Condition<Field, Variable>(Field.ASPECT, Operator.MATCHES, variable));
		return this;
	}

	@Override
	public QueryBuilder hasAnyAspectOf(final QName... aspects) {
		Assert.notEmpty(aspects, "Aspects cannot be empty.");
		for (int i = 0, n = aspects.length; i < n; i++) {
			final QName aspect = aspects[i];
			Assert.notNull(aspect, "Aspect cannot be null.");
			addCondition(new Condition<Field, QName>(Field.ASPECT, Operator.MATCHES, aspect));
			if (i < n - 1) {
				or();
			}
		}
		return this;
	}

	@Override
	public QueryBuilder hasAllAspectsOf(final QName... aspects) {
		Assert.notEmpty(aspects, "Aspects cannot be empty.");
		for (int i = 0, n = aspects.length; i < n; i++) {
			final QName aspect = aspects[i];
			Assert.notNull(aspect, "Aspect cannot be null.");
			addCondition(new Condition<Field, QName>(Field.ASPECT, Operator.MATCHES, aspect));
			if (i < n - 1) {
				and();
			}
		}
		return this;
	}

	@Override
	public QueryBuilder and() {
		Assert.state(getQuery().isEmpty() == false, "No conditions specified yet, cannot use logical operation.");
		Assert.state(logicalOperation == null,
				"Logical operation is already specified. Make sure you specify only one operation between conditions.");
		logicalOperation = LogicalOperation.AND;
		return this;
	}

	@Override
	public QueryBuilder or() {
		Assert.state(getQuery().isEmpty() == false, "No conditions specified yet, cannot use logical operation.");
		Assert.state(logicalOperation == null,
				"Logical operation is already specified. Make sure you specify only one operation between conditions.");
		logicalOperation = LogicalOperation.OR;
		return this;
	}

	@Override
	public QueryBuilder startSubquery() {
		final QueryBuilderImpl subqueryBuilder = new QueryBuilderImpl(this);
		addCondition(subqueryBuilder);
		return subqueryBuilder;
	};

	@Override
	public QueryBuilder endSubquery() {
		Assert.state(this.getParent() != null,
				"No parent query found. Make sure the numbers of calls to endSubquery() match that of calls to startSubquery().");
		return this.getParent();
	}

	@Override
	public QueryBuilder subquery(final Query subquery) {
		Assert.notNull(subquery, "Subquery cannot be null.");
		addCondition(subquery);
		return this;
	}

	@Override
	public QueryBuilder subquery(final QueryBuilder subquery) {
		Assert.notNull(subquery, "Subquery cannot be null.");
		addCondition(subquery);
		return this;
	}

}
